import yt

import yt_idv

ds = yt.load_sample("MOOSE_Sample_data", "out.e-s010")

rc = yt_idv.render_context(height=800, width=800, gui=True)

from yt_idv.cameras.trackball_camera import TrackballCamera  # NOQA
from yt_idv.scene_components.mesh import MeshRendering  # NOQA
from yt_idv.scene_data.mesh import MeshData  # NOQA
from yt_idv.scene_graph import SceneGraph  # NOQA

c = TrackballCamera(position=[3.5, 3.5, 3.5], focus=[0.0, 0.0, 0.0])
rc.scene = SceneGraph(camera=c)

dd = ds.all_data()
md = MeshData(data_source=dd)
md.add_data(("connect1", "diffused"))
mr = MeshRendering(data=md, cmap_log=False)

rc.scene.data_objects.append(md)
rc.scene.components.append(mr)
rc.run()



"""Top-level package for interactive volume rendering for yt."""

__author__ = """Matthew Turk"""
__email__ = "matthewturk@gmail.com"
__version__ = "0.2.3"

import os

# We don't want to import this if we're just doing offscreen rendering
from .rendering_contexts import render_context



#!/usr/bin/env python

"""Tests for `yt_idv` package."""

import base64

import pytest
import yt
from pytest_html import extras

import yt_idv


@pytest.fixture(autouse=True)
def pyopengl_setup(monkeypatch):
    monkeypatch.setenv("PYOPENGL_PLATFORM", "osmesa")


@pytest.fixture()
def osmesa_fake_amr():
    """Return an OSMesa context that has a "fake" AMR dataset added, with "radius"
    as the field.
    """
    ds = yt.testing.fake_amr_ds()
    dd = ds.all_data()
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    rc.add_scene(dd, "radius", no_ghost=True)
    yield rc
    rc.osmesa.OSMesaDestroyContext(rc.context)


@pytest.fixture()
def osmesa_empty():
    """Return an OSMesa context that has no dataset.
    """
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    ds = yt.testing.fake_amr_ds()
    rc.add_scene(ds, None)
    rc.ds = ds
    yield rc
    rc.osmesa.OSMesaDestroyContext(rc.context)


@pytest.fixture()
def image_store(request, extra, tmpdir):
    def _snap_image(rc):
        image = rc.run()
        img = yt.write_bitmap(image, None)
        content = base64.b64encode(img).decode("ascii")
        extra.append(extras.png(content))
        extra.append(extras.html("<br clear='all'/>"))

    return _snap_image


def test_snapshots(osmesa_fake_amr, image_store):
    """Check that we can make some snapshots."""
    osmesa_fake_amr.scene.components[0].render_method = "max_intensity"
    image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0].render_method = "projection"
    image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0].render_method = "transfer_function"
    image_store(osmesa_fake_amr)


def test_annotate_boxes(osmesa_empty, image_store):
    """Check the box annotation."""
    osmesa_empty.scene.add_box([0.0, 0.0, 0.0], [1.0, 1.0, 1.0])
    image_store(osmesa_empty)
    osmesa_empty.scene.add_box([0.2, 0.2, 0.3], [0.8, 0.8, 0.7])
    image_store(osmesa_empty)
    osmesa_empty.scene.annotations[-1].box_width /= 2
    osmesa_empty.scene.annotations[-1].box_color = (1.0, 0.0, 0.0)
    image_store(osmesa_empty)


def test_annotate_grids(osmesa_empty, image_store):
    """Make sure we can add some grid positions."""
    from yt_idv.scene_annotations.grid_outlines import GridOutlines  # NOQA
    from yt_idv.scene_data.grid_positions import GridPositions  # NOQA

    gp = GridPositions(grid_list=osmesa_empty.ds.index.grids.tolist())
    osmesa_empty.scene.data_objects.append(gp)
    go = GridOutlines(data=gp)
    osmesa_empty.scene.components.append(go)
    image_store(osmesa_empty)
    osmesa_empty.scene.camera.offset_position(0.25)
    image_store(osmesa_empty)
    osmesa_empty.scene.camera.offset_position(0.5)
    image_store(osmesa_empty)


def test_annotate_text(osmesa_empty, image_store):
    """Test that text can be annotated and updated."""
    text = osmesa_empty.scene.add_text("Origin 0 0", origin=(0.0, 0.0))
    image_store(osmesa_empty)
    text.text = "Change text"
    image_store(osmesa_empty)
    text.text = "Origin -0.5 -0.5"
    text.origin = (-0.5, -0.5)
    image_store(osmesa_empty)
    text.origin = (0.0, 0.0)
    text.text = "S 1.0"
    image_store(osmesa_empty)
    text.text = "S 2.0"
    text.scale = 2.0
    image_store(osmesa_empty)



import numpy as np
from OpenGL import GL, osmesa

from .base_offscreen import OffscreenRenderingContext


class OSMesaRenderingContext(OffscreenRenderingContext):
    """Rendering context using OSMesa (experimental)"""

    def __init__(self, width=1024, height=1024, **kwargs):
        super(OSMesaRenderingContext, self).__init__(width, height, **kwargs)
        self.osmesa = osmesa
        # Now we create our necessary bits.
        config_attribs = np.array(
            [
                osmesa.OSMESA_DEPTH_BITS,
                24,
                osmesa.OSMESA_STENCIL_BITS,
                8,
                osmesa.OSMESA_FORMAT,
                osmesa.OSMESA_RGBA,
                osmesa.OSMESA_PROFILE,
                osmesa.OSMESA_CORE_PROFILE,
                0,
            ],
            dtype="i4",
        )
        self.context = osmesa.OSMesaCreateContextAttribs(config_attribs, None)
        self._buffer = np.zeros((self.height, self.width, 4), dtype="u1")
        osmesa.OSMesaMakeCurrent(
            self.context, self._buffer, GL.GL_UNSIGNED_BYTE, self.height, self.width
        )

        GL.glClearColor(0.0, 0.0, 0.0, 0.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)



import traitlets
from OpenGL import GL

from yt_idv.constants import FULLSCREEN_QUAD
from yt_idv.opengl_support import (
    ColormapTexture,
    Framebuffer,
    VertexArray,
    VertexAttribute,
)
from yt_idv.scene_data.base_data import SceneData
from yt_idv.shader_objects import (
    ShaderProgram,
    ShaderTrait,
    component_shaders,
    default_shader_combos,
)

_cmaps = ["arbre", "viridis", "magma", "doom"]
_buffers = ["frame", "depth"]


class SceneComponent(traitlets.HasTraits):
    data = traitlets.Instance(SceneData)
    base_quad = traitlets.Instance(SceneData)
    name = "undefined"
    priority = traitlets.CInt(0)
    visible = traitlets.Bool(True)
    use_db = traitlets.Bool(False)  # use depth buffer
    last_use_db = traitlets.Bool(False)
    display_bounds = traitlets.Tuple(
        traitlets.CFloat(),
        traitlets.CFloat(),
        traitlets.CFloat(),
        traitlets.CFloat(),
        default_value=(0.0, 1.0, 0.0, 1.0),
    )
    clear_region = traitlets.Bool(False)

    render_method = traitlets.Unicode(allow_none=True)
    fragment_shader = ShaderTrait(allow_none=True).tag(shader_type="fragment")
    geometry_shader = ShaderTrait(allow_none=True).tag(shader_type="geometry")
    vertex_shader = ShaderTrait(allow_none=True).tag(shader_type="vertex")
    fb = traitlets.Instance(Framebuffer)
    colormap_fragment = ShaderTrait(allow_none=True).tag(shader_type="fragment")
    colormap_vertex = ShaderTrait(allow_none=True).tag(shader_type="vertex")
    colormap = traitlets.Instance(ColormapTexture)
    _program1 = traitlets.Instance(ShaderProgram, allow_none=True)
    _program2 = traitlets.Instance(ShaderProgram, allow_none=True)
    _program1_invalid = True
    _program2_invalid = True

    # These attributes are
    cmap_min = traitlets.CFloat(None, allow_none=True)
    cmap_max = traitlets.CFloat(None, allow_none=True)
    cmap_log = traitlets.Bool(True)
    scale = traitlets.CFloat(1.0)

    @traitlets.observe("display_bounds")
    def _change_display_bounds(self, change):
        # We need to update the framebuffer if the width or height has changed
        # Same thing is true if the total pixel size has changed, but that is
        # not doable from in here.
        if change["old"] == traitlets.Undefined:
            return
        old_width = change["old"][1] - change["old"][0]
        old_height = change["old"][3] - change["old"][2]
        new_width = change["new"][1] - change["new"][0]
        new_height = change["new"][3] - change["new"][2]
        if old_width != new_width or old_height != new_height:
            self.fb = Framebuffer()

    def render_gui(self, imgui, renderer, scene):
        changed, self.visible = imgui.checkbox("Visible", self.visible)
        changed, self.use_db = imgui.checkbox("Depth Buffer", self.use_db)
        if imgui.button("Recompile Shader"):
            shaders = (
                "vertex_shader",
                "geometry_shader",
                "fragment_shader",
                "colormap_vertex",
                "colormap_fragment",
            )
            for shader_name in shaders:
                s = getattr(self, shader_name, None)
                if s:
                    s.delete_shader()
            self._program1_invalid = self._program2_invalid = True
            changed = True
        _, cmap_index = imgui.listbox(
            "Colormap", _cmaps.index(self.colormap.colormap_name), _cmaps
        )
        if _:
            self.colormap.colormap_name = _cmaps[cmap_index]
        changed = changed or _
        _, self.cmap_log = imgui.checkbox("Take log", self.cmap_log)
        changed = changed or _
        if imgui.button("Reset Colorbounds"):
            self.cmap_min = self.cmap_max = None
            changed = True

        return changed

    @traitlets.default("render_method")
    def _default_render_method(self):
        return default_shader_combos[self.name]

    @traitlets.observe("render_method")
    def _change_render_method(self, change):
        new_combo = component_shaders[self.name][change["new"]]
        with self.hold_trait_notifications():
            self.vertex_shader = new_combo["first_vertex"]
            self.fragment_shader = new_combo["first_fragment"]
            self.geometry_shader = new_combo.get("first_geometry", None)
            self.colormap_vertex = new_combo["second_vertex"]
            self.colormap_fragment = new_combo["second_fragment"]

    @traitlets.default("fb")
    def _fb_default(self):
        return Framebuffer()

    @traitlets.observe("fragment_shader")
    def _change_fragment(self, change):
        # Even if old/new are the same
        self._program1_invalid = True

    @traitlets.observe("vertex_shader")
    def _change_vertex(self, change):
        # Even if old/new are the same
        self._program1_invalid = True

    @traitlets.observe("geometry_shader")
    def _change_geometry(self, change):
        self._program1_invalid = True

    @traitlets.observe("colormap_vertex")
    def _change_colormap_vertex(self, change):
        # Even if old/new are the same
        self._program2_invalid = True

    @traitlets.observe("colormap_fragment")
    def _change_colormap_fragment(self, change):
        # Even if old/new are the same
        self._program2_invalid = True

    @traitlets.default("colormap")
    def _default_colormap(self):
        cm = ColormapTexture()
        cm.colormap_name = "arbre"
        return cm

    @traitlets.default("vertex_shader")
    def _vertex_shader_default(self):
        return component_shaders[self.name][self.render_method]["first_vertex"]

    @traitlets.default("fragment_shader")
    def _fragment_shader_default(self):
        return component_shaders[self.name][self.render_method]["first_fragment"]

    @traitlets.default("geometry_shader")
    def _geometry_shader_default(self):
        _ = component_shaders[self.name][self.render_method]
        return _.get("first_geometry", None)

    @traitlets.default("colormap_vertex")
    def _colormap_vertex_default(self):
        return component_shaders[self.name][self.render_method]["second_vertex"]

    @traitlets.default("colormap_fragment")
    def _colormap_fragment_default(self):
        return component_shaders[self.name][self.render_method]["second_fragment"]

    @traitlets.default("base_quad")
    def _default_base_quad(self):
        bq = SceneData(
            name="fullscreen_quad", vertex_array=VertexArray(name="tri", each=6),
        )
        fq = FULLSCREEN_QUAD.reshape((6, 3), order="C")
        bq.vertex_array.attributes.append(
            VertexAttribute(name="vertexPosition_modelspace", data=fq)
        )
        return bq

    @property
    def program1(self):
        if self._program1_invalid:
            if self._program1 is not None:
                self._program1.delete_program()
            self._fragment_shader_default()
            self._program1 = ShaderProgram(
                self.vertex_shader, self.fragment_shader, self.geometry_shader
            )
            self._program1_invalid = False
        return self._program1

    @property
    def program2(self):
        if self._program2_invalid:
            if self._program2 is not None:
                self._program2.delete_program()
            # The vertex shader will always be the same.
            # The fragment shader will change based on whether we are
            # colormapping or not.
            self._program2 = ShaderProgram(self.colormap_vertex, self.colormap_fragment)
            self._program2_invalid = False
        return self._program2

    def run_program(self, scene):
        # Store this info, because we need to render into a framebuffer that is the
        # right size.
        x0, y0, w, h = GL.glGetIntegerv(GL.GL_VIEWPORT)
        GL.glViewport(0, 0, w, h)
        if not self.visible:
            return
        with self.fb.bind(True):
            with self.program1.enable() as p:
                scene.camera._set_uniforms(scene, p)
                self._set_uniforms(scene, p)
                with self.data.vertex_array.bind(p):
                    self.draw(scene, p)
        if (
            self.cmap_min is None
            or self.cmap_max is None
            or self.last_use_db != self.use_db
        ):
            self.last_use_db = self.use_db
            data = self.fb.data
            if self.use_db:
                data[:, :, :3] = self.fb.depth_data[:, :, None]
            data = data[data[:, :, 3] > 0][:, 0]
            if data.size > 0:
                self.cmap_min = cmap_min = data.min()
                self.cmap_max = cmap_max = data.max()
            if data.size == 0:
                cmap_min = 0.0
                cmap_max = 1.0
            else:
                print(f"Computed new cmap values {cmap_min} - {cmap_max}")
        else:
            cmap_min = float(self.cmap_min)
            cmap_max = float(self.cmap_max)
        with self.colormap.bind(0):
            with self.fb.input_bind(1, 2):
                with self.program2.enable() as p2:
                    with scene.bind_buffer():
                        p2._set_uniform("cmap", 0)
                        p2._set_uniform("fb_tex", 1)
                        p2._set_uniform("db_tex", 2)
                        p2._set_uniform("use_db", self.use_db)
                        # Note that we use cmap_min/cmap_max, not
                        # self.cmap_min/self.cmap_max.
                        p2._set_uniform("cmap_min", cmap_min)
                        p2._set_uniform("cmap_max", cmap_max)
                        p2._set_uniform("cmap_log", float(self.cmap_log))
                        with self.base_quad.vertex_array.bind(p2):
                            # Now we do our viewport globally, not just within
                            # the framebuffer
                            GL.glViewport(x0, y0, w, h)
                            GL.glDrawArrays(GL.GL_TRIANGLES, 0, 6)

    def draw(self, scene, program):
        raise NotImplementedError



from collections import defaultdict

import numpy as np
import traitlets
from yt.data_objects.data_containers import YTDataContainer

from yt_idv.opengl_support import Texture3D, VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class BlockCollection(SceneData):
    name = "block_collection"
    data_source = traitlets.Instance(YTDataContainer)
    texture_objects = traitlets.Dict(value_trait=traitlets.Instance(Texture3D))
    bitmap_objects = traitlets.Dict(value_trait=traitlets.Instance(Texture3D))
    blocks = traitlets.Dict(default_value=())
    scale = traitlets.Bool(False)
    blocks_by_grid = traitlets.Instance(defaultdict, (list,))
    grids_by_block = traitlets.Dict(default_value=())

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="block_info", each=1)

    def add_data(self, field, no_ghost=False):
        r"""Adds a source of data for the block collection.

        Given a `data_source` and a `field` to populate from, adds the data
        to the block collection so that is able to be rendered.

        Parameters
        ----------
        data_source : YTRegion
            A YTRegion object to use as a data source.
        field : string
            A field to populate from.
        no_ghost : bool (False)
            Should we speed things up by skipping ghost zone generation?
        """
        self.data_source.tiles.set_fields([field], [False], no_ghost=no_ghost)
        # Every time we change our data source, we wipe all existing ones.
        # We now set up our vertices into our current data source.
        vert, dx, le, re = [], [], [], []
        self.min_val = +np.inf
        self.max_val = -np.inf
        if self.scale:
            left_min = np.ones(3, "f8") * np.inf
            right_max = np.ones(3, "f8") * -np.inf
            for block in self.data_source.tiles.traverse():
                np.minimum(left_min, block.LeftEdge, left_min)
                np.maximum(right_max, block.LeftEdge, right_max)
            scale = right_max.max() - left_min.min()
            for block in self.data_source.tiles.traverse():
                block.LeftEdge -= left_min
                block.LeftEdge /= scale
                block.RightEdge -= left_min
                block.RightEdge /= scale
        for i, block in enumerate(self.data_source.tiles.traverse()):
            self.min_val = min(self.min_val, np.nanmin(np.abs(block.my_data[0])).min())
            self.max_val = max(self.max_val, np.nanmax(np.abs(block.my_data[0])).max())
            self.blocks[id(block)] = (i, block)
            vert.append([1.0, 1.0, 1.0, 1.0])
            dds = (block.RightEdge - block.LeftEdge) / block.source_mask.shape
            dx.append(dds.tolist())
            le.append(block.LeftEdge.tolist())
            re.append(block.RightEdge.tolist())
        for (g, node, (sl, _dims, _gi)) in self.data_source.tiles.slice_traverse():
            block = node.data
            self.blocks_by_grid[g.id - g._id_offset].append((id(block), i))
            self.grids_by_block[id(node.data)] = (g.id - g._id_offset, sl)

        if hasattr(self.min_val, "in_units"):
            self.min_val = self.min_val.d
        if hasattr(self.max_val, "in_units"):
            self.max_val = self.max_val.d

        LE = np.array([b.LeftEdge for i, b in self.blocks.values()]).min(axis=0)
        RE = np.array([b.RightEdge for i, b in self.blocks.values()]).max(axis=0)
        self.diagonal = np.sqrt(((RE - LE) ** 2).sum())
        # Now we set up our buffer
        vert = np.array(vert, dtype="f4")
        dx = np.array(dx, dtype="f4")
        le = np.array(le, dtype="f4")
        re = np.array(re, dtype="f4")

        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=vert)
        )
        self.vertex_array.attributes.append(VertexAttribute(name="in_dx", data=dx))
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_left_edge", data=le)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_right_edge", data=re)
        )

        # Now we set up our textures
        self._load_textures()

    def viewpoint_iter(self, camera):
        for block in self.data_source.tiles.traverse(viewpoint=camera.position):
            vbo_i, _ = self.blocks[id(block)]
            yield (vbo_i, self.texture_objects[vbo_i], self.bitmap_objects[vbo_i])

    def filter_callback(self, callback):
        # This is not efficient.  It calls it once for each node in a grid.
        # We do this the slow way because of the problem of ordering the way we
        # iterate over the grids and nodes.  This can be fixed at some point.
        for g_ind in self.blocks_by_grid:
            blocks = self.blocks_by_grid[g_ind]
            # Does this need an offset?
            grid = self.data_source.index.grids[g_ind]
            new_bitmap = callback(grid).astype("uint8")
            for b_id, _ in blocks:
                _, sl = self.grids_by_block[b_id]
                vbo_i, _ = self.blocks[b_id]
                self.bitmap_objects[vbo_i].data = new_bitmap[sl]

    def _load_textures(self):
        for block_id in sorted(self.blocks):
            vbo_i, block = self.blocks[block_id]
            n_data = np.abs(block.my_data[0]).copy(order="F").astype("float32").d
            # Avoid setting to NaNs
            if self.max_val != self.min_val:
                n_data = (n_data - self.min_val) / (
                    (self.max_val - self.min_val)
                )  # * self.diagonal)
            data_tex = Texture3D(data=n_data)
            bitmap_tex = Texture3D(
                data=block.source_mask * 255, min_filter="nearest", mag_filter="nearest"
            )
            self.texture_objects[vbo_i] = data_tex
            self.bitmap_objects[vbo_i] = bitmap_tex



import numpy as np
import pyglet
from yt import write_bitmap

from yt_idv.simple_gui import SimpleGUI

from .base_context import BaseContext


class PygletRenderingContext(pyglet.window.Window, BaseContext):
    """Basic rendering context for IDV using GLFW3, that handles the main window even loop

    Parameters
    ----------
    width : int, optional
        The width of the Interactive Data Visualization window.  For
        performance reasons it is recommended to use values that are natural
        powers of 2.
    height : int, optional
        The height of the Interactive Data Visualization window.  For
        performance reasons it is recommended to use values that are natural
        powers of 2.
    title : str, optional
        The title of the Interactive Data Visualization window.
    position : tuple of ints, optional
        What position should the window be moved to? (Upper left)  If not
        specified, default to center.
    """

    _do_update = True

    def __init__(
        self,
        width=1024,
        height=1024,
        title="vol_render",
        position=None,
        visible=True,
        gui=True,
        scene=None,
    ):
        self.offscreen = not visible
        config = pyglet.gl.Config(
            major_version=3,
            minor_version=3,
            forward_compat=True,
            double_buffer=True,
            depth_size=24,
        )
        super(PygletRenderingContext, self).__init__(
            width, height, config=config, visible=visible, caption=title
        )
        if position is None:
            self.center_window()
        else:
            # self.set_position(*position)
            self.set_location(*position)

        if gui:
            gui = SimpleGUI(self)
        self.gui = gui
        self.scene = scene

    def on_draw(self):
        self.switch_to()
        if self._do_update and self.scene:
            self._do_update = False
            self.clear()
            if self.scene is not None:
                # This should check if the scene is actually dirty, and only
                # redraw the FB if it's not; this might need another flag
                self.scene.render()
                if self.image_widget is not None:
                    self.image_widget.value = write_bitmap(
                        self.scene.image[:, :, :3], None
                    )
        if self.gui:
            self.switch_to()
            self.gui.render(self.scene)

    def set_position(self, xpos, ypos):
        if xpos < 0 or ypos < 0:
            raise RuntimeError
        max_width = self.screen.width
        max_height = self.screen.height
        win_width, win_height = self.width, self.height
        if 0 < xpos < 1:
            # We're being fed relative coords.  We offset these for the window
            # center.
            xpos = max(self.screen.x + xpos * max_width - 0.5 * win_width, 0)
        if 0 < ypos < 1:
            # We're being fed relative coords.  We offset these for the window
            # center.
            ypos = max(self.screen.y + ypos * max_height - 0.5 * win_height, 0)
        print("Setting position", xpos, ypos)
        self.set_location(int(xpos), int(ypos))

    def center_window(self):
        self.set_position(0.5, 0.5)

    def on_mouse_press(self, x, y, button, modifiers):
        self._do_update = True

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        if self.gui and self.gui.mouse_event_handled:
            self._do_update = True
            return
        start_x = -1.0 + 2.0 * x / self.width
        end_x = -1.0 + 2.0 * (x - dx) / self.width
        start_y = -1.0 + 2.0 * y / self.height
        end_y = -1.0 + 2.0 * (y + dy) / self.height

        self.scene.camera.update_orientation(start_x, start_y, end_x, end_y)
        self._do_update = True

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        # captures mouse scrolling as zoom in/out
        if self.gui and self.gui.mouse_event_handled:
            self._do_update = True
            return

        camera = self.scene.camera  # current camera
        dpos = (
            0.1
            * (camera.position - camera.focus)
            / np.linalg.norm(camera.position - camera.focus)
        )

        # wheel scroll comes in the scroll_y parameter with a value +/- 1
        # +1 when scrolling "down", -1 when scrolling "up", so
        # flip it so scrolling "down" zooms out:
        zoom_inout = -1 * scroll_y
        self.scene.camera.offset_position(zoom_inout * dpos)
        self._do_update = True

    def on_key_press(self, symbol, modifiers):
        # skeleton for capturing key presses!

        # potential navigation keys
        if symbol in [pyglet.window.key.LEFT, pyglet.window.key.A]:
            pass
        if symbol in [pyglet.window.key.RIGHT, pyglet.window.key.D]:
            pass
        if symbol in [pyglet.window.key.UP, pyglet.window.key.W]:
            pass
        if symbol in [pyglet.window.key.DOWN, pyglet.window.key.S]:
            pass

    def run(self):
        pyglet.app.run()



from math import ceil, floor

import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.opengl_support import TransferFunctionTexture
from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.shader_objects import component_shaders


class BlockRendering(SceneComponent):
    """
    A class that renders block data.  It may do this in one of several ways,
    including mesh outline.  This allows us to render a single collection of
    blocks multiple times in a single scene and to separate out the memory
    handling from the display.
    """

    name = "block_rendering"
    data = traitlets.Instance(BlockCollection)
    box_width = traitlets.CFloat(0.1)
    sample_factor = traitlets.CFloat(1.0)
    transfer_function = traitlets.Instance(TransferFunctionTexture)
    tf_min = traitlets.CFloat(0.0)
    tf_max = traitlets.CFloat(1.0)
    tf_log = traitlets.Bool(True)

    priority = 10

    def render_gui(self, imgui, renderer, scene):
        changed = super(BlockRendering, self).render_gui(imgui, renderer, scene)
        _, sample_factor = imgui.slider_float(
            "Sample Factor", self.sample_factor, 1.0, 20.0
        )
        if _:
            self.sample_factor = sample_factor
        # Now, shaders
        shader_combos = list(sorted(component_shaders[self.name]))
        descriptions = [
            component_shaders[self.name][_]["description"] for _ in shader_combos
        ]
        selected = shader_combos.index(self.render_method)
        _, shader_ind = imgui.listbox("Shader", selected, descriptions)
        if _:
            self.render_method = shader_combos[shader_ind]
        changed = changed or _
        if imgui.button("Add Block Outline"):
            from ..scene_annotations.block_outline import BlockOutline

            block_outline = BlockOutline(data=self.data)
            scene.annotations.append(block_outline)
        if imgui.button("Add Grid Outline"):
            from ..scene_annotations.grid_outlines import GridOutlines
            from ..scene_data.grid_positions import GridPositions

            grids = self.data.data_source.ds.index.grids.tolist()
            gp = GridPositions(grid_list=grids)
            scene.data_objects.append(gp)
            scene.components.append(GridOutlines(data=gp))
        if self.render_method == "transfer_function":
            # Now for the transfer function stuff
            imgui.image_button(
                self.transfer_function.texture_name, 256, 32, frame_padding=0
            )
            imgui.text("Right click and drag to change")
            update = False
            data = self.transfer_function.data.astype("f4") / 255
            for i, c in enumerate("rgba"):
                imgui.plot_lines(
                    f"## {c}",
                    data[:, 0, i].copy(),
                    scale_min=0.0,
                    scale_max=1.0,
                    graph_size=(256, 32),
                )
                if imgui.is_item_hovered() and imgui.is_mouse_dragging(2):
                    update = True
                    dx, dy = renderer.io.mouse_delta
                    dy = -dy
                    mi = imgui.get_item_rect_min()
                    ma = imgui.get_item_rect_max()
                    x, y = renderer.io.mouse_pos
                    x = x - mi.x
                    y = (ma.y - mi.y) - (y - mi.y)
                    xb1 = floor(min(x + dx, x) * data.shape[0] / (ma.x - mi.x))
                    xb2 = ceil(max(x + dx, x) * data.shape[0] / (ma.x - mi.x))
                    yv1 = y / (ma.y - mi.y)
                    yv2 = (y + dy) / (ma.y - mi.y)
                    yv1, yv2 = (max(min(_, 1.0), 0.0) for _ in (yv1, yv2))
                    if dx < 0:
                        yv2, yv1 = yv1, yv2
                        xb1 -= 1
                    elif dx > 0:
                        xb2 += 1
                    xb1 = max(0, xb1)
                    xb2 = min(255, xb2)
                    if renderer.io.key_shift:
                        yv1 = yv2 = 1.0
                    elif renderer.io.key_ctrl:
                        yv1 = yv2 = 0.0
                    data[xb1:xb2, 0, i] = np.mgrid[yv1 : yv2 : (xb2 - xb1) * 1j]
            if update:
                self.transfer_function.data = (data * 255).astype("u1")
        return changed

    @traitlets.default("transfer_function")
    def _default_transfer_function(self):
        tf = TransferFunctionTexture(data=np.ones((256, 1, 4), dtype="u1") * 255)
        return tf

    def draw(self, scene, program):
        each = self.data.vertex_array.each
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        with self.transfer_function.bind(target=2):
            for tex_ind, tex, bitmap_tex in self.data.viewpoint_iter(scene.camera):
                with tex.bind(target=0):
                    with bitmap_tex.bind(target=1):
                        GL.glDrawArrays(GL.GL_POINTS, tex_ind * each, each)

    def _set_uniforms(self, scene, shader_program):
        shader_program._set_uniform("box_width", self.box_width)
        shader_program._set_uniform("sample_factor", self.sample_factor)
        shader_program._set_uniform("ds_tex", 0)
        shader_program._set_uniform("bitmap_tex", 1)
        shader_program._set_uniform("tf_tex", 2)
        shader_program._set_uniform("tf_min", self.tf_min)
        shader_program._set_uniform("tf_max", self.tf_max)
        shader_program._set_uniform("tf_log", float(self.tf_log))



import math

import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.particle_positions import ParticlePositions


class ParticleRendering(SceneComponent):
    name = "particle_rendering"
    data = traitlets.Instance(ParticlePositions)
    scale = traitlets.CFloat(1e-3)
    max_particle_size = traitlets.CFloat(1e-3)

    def render_gui(self, imgui, renderer, scene):
        changed = super(ParticleRendering, self).render_gui(imgui, renderer, scene)
        _, new_value = imgui.slider_float(
            "Log Scale", math.log10(self.scale), -8.0, 2.0
        )
        if _:
            self.scale = 10 ** new_value
            changed = True
        imgui.text("Filter Particle Max Size")
        _, new_value = imgui.slider_float("", 1.0 / self.max_particle_size, 1.0, 100.0)
        if _:
            self.max_particle_size = 1.0 / new_value
            changed = True
        return changed

    def draw(self, scene, program):
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        GL.glDrawArraysInstanced(GL.GL_TRIANGLE_STRIP, 0, 4, self.data.size)

    def _set_uniforms(self, scene, shader_program):
        cam = scene.camera
        shader_program._set_uniform("scale", self.scale)
        shader_program._set_uniform("projection", cam.projection_matrix)
        shader_program._set_uniform("modelview", cam.view_matrix)
        shader_program._set_uniform("max_particle_size", self.max_particle_size)
        shader_program._set_uniform(
            "inv_pmvm", np.linalg.inv(cam.projection_matrix @ cam.view_matrix)
        )



import yt

import yt_idv

N = 400

ds = yt.load_sample("HiresIsolatedGalaxy")
c = [0.53, 0.53, 0.53]

rc = yt_idv.render_context("egl", width=1024, height=1024)
sc = rc.add_scene(ds, "density", no_ghost=False)
sc.components[0].render_method = "projection"
sc.camera.focus = c
sc.camera.position = [0.45, 0.44, 0.43]

ds = (sc.camera.focus - sc.camera.position) / N

for i in range(N):
    sc.components[0].cmap_min = sc.components[0].cmap_max = None
    sc.camera.position = sc.camera.position + ds
    sc.camera._update_matrices()
    rc.snap()



import numpy as np
import OpenGL.GL as GL
import traitlets
from yt.data_objects.data_containers import YTDataContainer
from yt.data_objects.index_subobjects.octree_subset import OctreeSubsetBlockSlice

from yt_idv.constants import aabb_triangle_strip
from yt_idv.opengl_support import Texture3D, VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class OctreeBlockCollection(SceneData):
    name = "octree_block_collection"
    data_source = traitlets.Instance(YTDataContainer)
    data_textures = traitlets.List(value_trait=traitlets.Instance(Texture3D))
    bitmap_textures = traitlets.List(value_trait=traitlets.Instance(Texture3D))
    shapes = traitlets.List(value_trait=traitlets.CInt())

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="octree_block_info", each=14)

    def add_data(self, field):
        r"""Adds a source of data for the block collection.

        Given a `data_source` and a `field` to populate from, adds the data
        to the block collection so that is able to be rendered.

        Parameters
        ----------
        data_source : YTRegion
            A YTRegion object to use as a data source.
        field : string
            A field to populate from.
        no_ghost : bool (False)
            Should we speed things up by skipping ghost zone generation?
        """
        ds = self.data_source.ds
        ds.index._identify_base_chunk(self.data_source)
        left_edges = []
        right_edges = []
        dx = []
        data = []

        for obj in self.data_source._current_chunk.objs:
            bs = OctreeSubsetBlockSlice(obj, ds)
            LE = bs._fcoords[0, 0, 0, :, :].d - bs._fwidth[0, 0, 0, :, :].d * 0.5
            RE = bs._fcoords[-1, -1, -1, :, :].d + bs._fwidth[-1, -1, -1, :, :].d * 0.5
            dx.append(bs._fwidth[-1, -1, -1, :, :].d)
            left_edges.append(LE)
            right_edges.append(RE)
            d = bs.get_vertex_centered_data([field])[field]
            data.append(
                np.concatenate(
                    [d[:, :, :, i] for i in range(d.shape[-1])], axis=2
                ).copy(order="C")
            )

        # Let's reshape ...
        left_edges = np.concatenate(left_edges, axis=0).astype("f4")
        right_edges = np.concatenate(right_edges, axis=0).astype("f4")
        dx = np.concatenate(dx, axis=0).astype("f4")
        data = np.concatenate(data, axis=-1).astype("f4")
        data = data.reshape((3, 3, -1))

        self.min_val = np.nanmin(data)
        self.max_val = np.nanmax(data)

        if hasattr(self.min_val, "in_units"):
            self.min_val = self.min_val.d
        if hasattr(self.max_val, "in_units"):
            self.max_val = self.max_val.d

        if self.max_val != self.min_val:
            data = (data - self.min_val) / (
                (self.max_val - self.min_val)
            )  # * self.diagonal)

        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=aabb_triangle_strip, divisor=0)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_dx", data=dx, divisor=1)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_left_edge", data=left_edges, divisor=1)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_right_edge", data=right_edges, divisor=1)
        )

        # Now we set up our textures; we need to break our texture up into
        # groups of MAX_3D_TEXTURE_SIZE
        tex_size = GL.glGetInteger(GL.GL_MAX_3D_TEXTURE_SIZE)
        # for now, use one texture for all the bitmaps
        bt = Texture3D(data=np.ones((3, 3, tex_size), dtype="u1") * 255)
        for start_index in np.mgrid[0 : data.shape[-1] : tex_size]:
            d = data[:, :, start_index : start_index + tex_size]
            self.data_textures.append(Texture3D(data=d))
            self.bitmap_textures.append(bt)
            self.shapes.append(d.shape[-1])



import yt

import yt_idv
from yt_idv.scene_components.particles import ParticleRendering
from yt_idv.scene_data.particle_positions import ParticlePositions

ds = yt.load_sample("IsolatedGalaxy")
dd = ds.all_data()

rc = yt_idv.render_context(height=800, width=800, gui=True)
sg = rc.add_scene(ds, None, no_ghost=True)

pos = ParticlePositions(data_source=dd)
pren = ParticleRendering(data=pos)

sg.data_objects.append(pos)
sg.components.append(pren)

rc.run()



import imgui
import matplotlib.pyplot as plt
import numpy as np
from imgui.integrations.pyglet import create_renderer
from yt.visualization.image_writer import write_bitmap, write_image

from .opengl_support import Texture2D


class SimpleGUI:

    renderer = None
    callbacks = None
    context = None
    window = None
    draw = False

    def __init__(self, window):
        self.window = window
        self.context = imgui.create_context()
        self.renderer = create_renderer(window)
        self.snapshot_count = 0
        self.snapshot_format = r"snap_{count:04d}.png"
        data = plt.get_cmap("viridis")(np.mgrid[0.0:1.0:256j]).reshape((-1, 1, 4))
        self.data = dict(
            r=data[:, 0, 0].astype("f4"),
            g=data[:, 0, 1].astype("f4"),
            b=data[:, 0, 2].astype("f4"),
            a=data[:, 0, 3].astype("f4"),
        )
        data = (data[:, :, :4] * 255).astype("u1")
        self.colormap = Texture2D(data=data, boundary_x="clamp", boundary_y="clamp")

    def render(self, scene):
        imgui.new_frame()
        changed = False
        if scene is not None:
            imgui.style_colors_classic()
            imgui.begin("Scene")
            imgui.text("Filename Template:")
            _, self.snapshot_format = imgui.input_text("", self.snapshot_format, 256)
            if imgui.button("Save Snapshot"):
                # Call render again, since we're in the middle of overlaying
                # some stuff and we want a clean scene snapshot
                scene.render()
                write_bitmap(
                    scene.image[:, :, :3],
                    self.snapshot_format.format(count=self.snapshot_count),
                )
                self.snapshot_count += 1
            if imgui.tree_node("Debug"):
                if imgui.button("Save Depth"):
                    scene.render()
                    write_image(
                        scene.depth,
                        self.snapshot_format.format(count=self.snapshot_count),
                    )
                    self.snapshot_count += 1
                imgui.tree_pop()
            _ = self.render_camera(scene)
            changed = changed or _
            # imgui.show_style_editor()
            for i, element in enumerate(scene):
                if imgui.tree_node(f"element {i + 1}: {element.name}"):
                    changed = changed or element.render_gui(imgui, self.renderer, scene)
                    imgui.tree_pop()
            self.window._do_update = self.window._do_update or changed
            imgui.end()
        imgui.render()
        self.renderer.render(imgui.get_draw_data())

    def render_camera(self, scene):
        if not imgui.tree_node("Camera"):
            return
        changed = False
        with scene.camera.hold_trait_notifications():
            for attr in ("position", "up", "focus"):
                arr = getattr(scene.camera, attr)
                imgui.text(f"Camera {attr}")
                _, values = imgui.input_float3(
                    "",
                    arr[0],
                    arr[1],
                    arr[2],
                    flags=imgui.INPUT_TEXT_ENTER_RETURNS_TRUE,
                )
                changed = changed or _
                if _:
                    setattr(scene.camera, attr, np.array(values))
            _, values = imgui.input_float2(
                "Camera Planes",
                scene.camera.near_plane,
                scene.camera.far_plane,
                format="%0.6f",
                flags=imgui.INPUT_TEXT_ENTER_RETURNS_TRUE,
            )
            changed = changed or _
            if _:
                scene.camera.near_plane = values[0]
                scene.camera.far_plane = values[1]
            if imgui.button("Center"):
                scene.camera.position = np.array([0.499, 0.499, 0.499])
                scene.camera.focus = np.array([0.5, 0.5, 0.5])
                changed = True
            if imgui.button("Outside"):
                scene.camera.position = np.array([1.5, 1.5, 1.5])
                scene.camera.focus = np.array([0.5, 0.5, 0.5])
                changed = True
        if changed:
            scene.camera._update_matrices()
        imgui.tree_pop()
        return changed

    @property
    def mouse_event_handled(self):
        return self.renderer.io.want_capture_mouse

    @property
    def keyboard_event_handled(self):
        return self.renderer.io.want_capture_keyboard



import numpy as np
import traitlets

from yt_idv.opengl_support import VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class LineData(SceneData):
    name = "line_data"
    n_values = traitlets.CInt()

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="vertices", each=6)

    def add_data(self, lines):
        assert lines.shape[1] == 4
        x_coord = np.mgrid[0.0 : 1.0 : lines.shape[0] * 1j].astype("f4")
        x_coord = x_coord.reshape((-1, 1))
        self.n_vertices = lines.shape[0]
        self.vertex_array.attributes.append(
            VertexAttribute(name="rgba_values", data=lines)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="x_coord", data=x_coord)
        )



import yt

import yt_idv
from yt_idv.scene_components.sph_particles import SPHRendering
from yt_idv.scene_data.particle_positions import ParticlePositions

ds = yt.load_sample("snapshot_033")
dd = ds.all_data()

rc = yt_idv.render_context(height=800, width=800, gui=True)
sg = rc.add_scene(ds, None, no_ghost=True)

pos = ParticlePositions(
    data_source=dd,
    radius_field="smoothing_length",
    particle_type="gas",
    color_field="density",
    position_field="position",
)
pren = SPHRendering(data=pos, scale=1.0, max_particle_size=1.0)

sg.data_objects.append(pos)
sg.components.append(pren)

print(sg.camera.near_plane, sg.camera.far_plane)
rc.run()



from yt import write_bitmap

from .base_context import BaseContext


class OffscreenRenderingContext(BaseContext):
    """Base class for offscreen rendering."""

    def run(self):

        if self.scene is None:
            return
        self.scene.render()
        if self.image_widget is not None:
            self.image_widget.value = write_bitmap(self.scene.image[:, :, :3], None)
            return
        return self.scene.image

    def snap(self, *args, **kwargs):
        if self.scene is None:
            return
        self.scene.render()
        super().snap(*args, **kwargs)



from ctypes import pointer

import numpy as np
from OpenGL import EGL, GL

from .base_offscreen import OffscreenRenderingContext


class EGLRenderingContext(OffscreenRenderingContext):
    """Rendering context using EGL (experimental)

    Parameters
    ----------
    width : int, optional
        The width of the off-screen buffer window.  For performance reasons it
        is recommended to use values that are natural powers of 2.

    height : int, optional
        The height of the off-screen buffer window.  For performance reasons it
        it is recommended to use values that are natural powers of 2.

    """

    def __init__(self, width=1024, height=1024, **kwargs):
        super(EGLRenderingContext, self).__init__(width, height, **kwargs)

        self.EGL = EGL
        self.display = EGL.eglGetDisplay(EGL.EGL_DEFAULT_DISPLAY)
        major = np.zeros(1, "i4")
        minor = np.zeros(1, "i4")
        EGL.eglInitialize(self.display, major, minor)
        num_configs = np.zeros(1, "i4")
        config = EGL.EGLConfig()
        # Now we create our necessary bits.
        config_attribs = np.array(
            [
                EGL.EGL_RED_SIZE,
                8,
                EGL.EGL_GREEN_SIZE,
                8,
                EGL.EGL_BLUE_SIZE,
                8,
                EGL.EGL_DEPTH_SIZE,
                24,
                EGL.EGL_STENCIL_SIZE,
                8,
                EGL.EGL_COLOR_BUFFER_TYPE,
                EGL.EGL_RGB_BUFFER,
                EGL.EGL_SURFACE_TYPE,
                EGL.EGL_PBUFFER_BIT,
                EGL.EGL_RENDERABLE_TYPE,
                EGL.EGL_OPENGL_BIT,
                EGL.EGL_CONFIG_CAVEAT,
                EGL.EGL_NONE,
                EGL.EGL_NONE,
            ],
            dtype="i4",
        )
        EGL.eglChooseConfig(
            self.display, config_attribs, pointer(config), 1, num_configs
        )

        pbuffer_attribs = np.array(
            [EGL.EGL_WIDTH, width, EGL.EGL_HEIGHT, height, EGL.EGL_NONE], dtype="i4"
        )
        self.surface = EGL.eglCreatePbufferSurface(
            self.display, config, pbuffer_attribs
        )
        EGL.eglBindAPI(EGL.EGL_OPENGL_API)

        self.context = EGL.eglCreateContext(
            self.display, config, EGL.EGL_NO_CONTEXT, None
        )

        EGL.eglMakeCurrent(self.display, self.surface, self.surface, self.context)

        GL.glClearColor(0.0, 0.0, 0.0, 0.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)



import numpy
from setuptools import Extension, find_packages, setup

from Cython.Build import cythonize  # isort:skip

setup(
    ext_modules=cythonize("yt_idv/*.pyx"),
    include_dirs=[numpy.get_include()],
)



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_annotations.base_annotation import SceneAnnotation
from yt_idv.scene_data.grid_positions import GridPositions


class GridOutlines(SceneAnnotation):
    """
    A class that renders outlines of grid positions.
    """

    name = "grid_outline"
    data = traitlets.Instance(GridPositions)
    box_width = traitlets.CFloat(0.25)  # quarter of a dx
    box_color = traitlets.Tuple((1.0, 1.0, 1.0), trait=traitlets.CFloat())
    box_alpha = traitlets.CFloat(1.0)

    def draw(self, scene, program):
        GL.glDisable(GL.GL_CULL_FACE)
        GL.glDrawArrays(GL.GL_POINTS, 0, len(self.data.grid_list))

    def _set_uniforms(self, scene, shader_program):
        shader_program._set_uniform("box_width", self.box_width)
        shader_program._set_uniform("box_color", np.array(self.box_color))
        shader_program._set_uniform("box_alpha", self.box_alpha)
